
/**************************************************************************
 *                                                                        *
 * Copyright (c) 1996 Michael Richmond and Richard Treffers               *
 *                                                                        *
 *                    This software may be copied and distributed for     *
 *                    educational, research and not for profit services   *
 *                    provided that this copyright and statement are      *
 *                    included in all such copies.                        *
 *                                                                        *
 **************************************************************************/


/*	
	attempts to remove "readout smear" from images taken without a 
   physical shutter.  Based on the ideas published in
      Owen, Weaver, & Cheng, PASP 132, 014504 (2020)
	
   4/29/2020 MWR
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "pcvista.h"
#include "fits.h"

#ifdef PROTO
int main(int, char **);
void desmear_col(int nrow, int ncol, double flushtime, double readtime);
#else
void desmear_col();
#endif

static int ncol;
static int nrow;
static FITS_HANDLE fin, fout;
static int16 data[NMAX], cor_data[NMAX];


#define DEBUG

	/* default time to flush CCD, in seconds */
#define DEFAULT_FLUSHTIME   0.001
	/* default time to read CCD, in seconds */
#define DEFAULT_READTIME   -1

#define USAGE "usage: desmear file_name  [flush=] [read=] [outfile=] [help]"

int
main(argc,argv)
int argc;
char *argv[];
{
	int i;
	char *gotit,outfile[NBUF];
	char buf[NBUF];
	int new_file=0;
	int naxis1 = -1;
	int naxis2 = -1;
	double flushtime = DEFAULT_FLUSHTIME;
	double readtime = DEFAULT_READTIME;

	if (argc < 2) {
		error(-1, USAGE);
	}
	if (strcmp(argv[1], "help") == 0) {
		printf("%s\n", USAGE);
		exit(0);
	}

	fin = fits_open(argv[1], "r", &nrow, &ncol);
	strcpy(outfile,argv[1]);

	for(i=2; i<argc; i++){
		if(keyword("help",argv[i])) {
			printf("%s\n",USAGE);
			exit(0);
		}
		if ((gotit = find("outfile", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			new_file=1;
			continue;
		}
		if ((gotit = find("flush", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			flushtime = evaluate(gotit);
			continue;
		}
		if ((gotit = find("read", argv[i])) != NULL) {
			strcpy(outfile,gotit);
			readtime = evaluate(gotit);
			continue;
		}
		fprintf(stderr,"desmear: unknown option '%s'\n",argv[i]);
		exit(1);
	}

	fout = fits_open(outfile, new_file?"w":"x", &nrow, &ncol);
		 
	if(new_file)
		fits_copyheader(fin,fout, FITS_CHECK);

	desmear_col(nrow, ncol, flushtime, readtime);
	
	fits_close(fout);
	exit(0);
}


	/*********************************************************************
	 * PROCEDURE: desmear_col
	 * 
	 * DESCRIPTION: Given the number of rows and cols in a FITS image,
	 *              and information on the flush and readout times,
	 *              we apply the corrections described in 
	 *              Owen, Weaver, and Chung.  
	 * 
	 *              If the readout time is set to -1, assume that it
	 *              is the same as the flush time.  In that case,
	 *              we only need to make the first set of corrections.
	 *
	 *              If the readout time is NOT the same as the flush time,
	 *              (and it is positive), then after having made the first
	 *              set of corrections for flush time, we make a second
	 *              round of corrections.
	 *               
	 *              The corrected data is written into the (global)
	 *              output FITS handle.
	 *
	 * RETURNS:     nothing
	 */

void
desmear_col(
	int nrow,           /* number of rows in image */
	int ncol,           /* number of columns in image */
	double flushtime,   /* time to flush chip pre-exposure (sec) */
	double readtime    /* time to read chip post-exposure (sec) */
)
{
	char buf[NBUF], err_msg[NBUF];
	int row, col;
	int second_pass_flag = 0;
	double col_sum[NMAX], delta[NMAX], second_cor[NMAX];
	double exptime, eps;
	double top, bottom;
	double val, cor_val;
	double rho;

#ifdef DEBUG
	printf("desmear: nrow            %5d \n", nrow);
	printf("desmear: ncol            %5d \n", ncol);
	printf("desmear: flush time      %14.8f \n", flushtime);
	printf("desmear: read time       %14.8f \n", readtime);
#endif

	if (fits_get_symbol(fin, "EXPTIME", buf) == FITS_PASS) {
		exptime = atof(buf);
#ifdef DEBUG
		printf(" read exptime = %.3f \n", exptime);
#endif
		if (exptime <= 0) {
			sprintf(err_msg, "desmear: exptime %f is <= 0 ?! \n", exptime);
			error(1, err_msg);
		}
	}


	/* step 1: set/compute the some constants we'll need later */
	eps = flushtime / exptime;
	if ((flushtime == readtime) || (readtime == -1)) {
		/* no need to compute rho, we won't be using it */
		second_pass_flag = 0;
		rho = 0.0;
	}
	else {
		/* we need this variable for second set of corrections */
		second_pass_flag = 1;
		rho = (readtime - flushtime) / exptime;
	}
#ifdef DEBUG
	printf("desmear: eps is %7.4f / %7.4f = %7.4f \n", flushtime, exptime, eps);
	printf("desmear: rho is (%7.4e - %7.4e) / %7.4e = %7.4e \n", 
					readtime, flushtime, exptime, rho);
#endif
	
	/* step 2: compute sum of pixel values along each column */
	for (col = 0; col < ncol; col++) {
		col_sum[col] = 0.0;
	}
	for (row = 0; row < nrow; row++) {
		fits_get_data(fin, row, 0, data, ncol);
		for (col = 0; col < ncol; col++) {
			col_sum[col] += (double) data[col];
		}
	}

	/* step 3: compute the excess signal due to flush transfer */
	bottom = 1.0 + (nrow - 1)*eps;
	for (col = 0; col < ncol; col++) {
		top = col_sum[col]*eps;
		delta[col] = top/bottom;
	}
#ifdef DEBUG
	printf(" bottom =   1.0 + (%5d - 1)*%.4f = %8.4f \n",
		nrow, eps, bottom);
	for (col = 0; col < ncol; col++) {
		printf("  col %5d  col_sum %12.0f  delta %12.5f \n", 
             col, col_sum[col], delta[col]);
	}
#endif

	/* step 4: compute the corrected pixel values */
	bottom = 1.0 - eps;
	for (row = 0; row < nrow; row++) {
		fits_get_data(fin, row, 0, data, ncol);
		for (col = 0; col < ncol; col++) {
			val = (double) data[col];
			top = val - delta[col];
			cor_val = (top / bottom);
			if (cor_val < MIN_DATA_VAL) {
				cor_val = MIN_DATA_VAL;
			}
			else if (cor_val > MAX_DATA_VAL) {
				cor_val = MAX_DATA_VAL;
			}
			cor_data[col] = (int16) cor_val;
		}

		/* 
		 * if necessary (flush and readout times not identical),
		 *    we make a second set of corrections here
		 */
		if (second_pass_flag == 1) {
			if (row == 0) {
				/* 
				 * special case, no extra modifications needed;
				 *     just prepare corrections for next row 
			 	 */	
				for (col = 0; col < ncol; col++) {
					second_cor[col] = rho*cor_data[col];
				}
			}
			else {
				for (col = 0; col < ncol; col++) {
					cor_data[col] -= second_cor[col];
					second_cor[col] += rho*cor_data[col];
				}
			}
		}
	
		fits_put_data(fout, row, 0, cor_data, ncol);
	}

	


#if 0
	for (row = 0; row < nrow; row++) {
		fits_get_data(fin, row, 0, data1, ncol);
		for (i = 0, j = ncol - 1; i < j; i++, j--) {
			temp = data1[i];
			data1[i] = data1[j];
			data1[j] = temp;
		}
		fits_put_data(fout, row, 0, data1, ncol);
	}
#endif

	return;
}
